CREATED_BY_LABEL = ("created_by.gefyra.dev", "true")
CARGO_ENDPOINT_LABEL = "cargo_endpoint.gefyra.dev"
VERSION_LABEL = "version.gefyra.dev"
ACTIVE_KUBECONFIG_LABEL = "active_kubeconfig.gefyra.dev"
ACTIVE_KUBECONFIG_CONTEXT_LABEL = "active_context.gefyra.dev"



import json
import logging
from pathlib import Path
from typing import Optional

logger = logging.getLogger("gefyra")

MINIKUBE_CONFIG = "~/.minikube/profiles/{profile}/config.json"


def _read_minikube_config(profile: str = "minikube") -> dict:
    config_file = Path(MINIKUBE_CONFIG.format(profile=profile)).expanduser()
    with open(config_file, "r") as f:
        data = json.load(f)
    return data


def _get_a_worker_ip(config: dict):
    nodes = config["Nodes"]
    for node in nodes:
        if node["Worker"]:
            return node["IP"]
    raise RuntimeError("This Minikube cluster does not have a worker node.")


def detect_minikube_config(profile: Optional[str]) -> dict:
    """
    Read the config for a local Minikube cluster from its configuration file and set Gefyra accordingly
    :return: a preped ClientConfiguration object
    """
    try:
        config = _read_minikube_config(profile)
    except FileNotFoundError:
        raise RuntimeError(
            f"The minikube profile {profile} does not exist. Did you start Minikube?"
            f"Please also review your profile with 'minikube profile list' and try again. "
            f"Minikube profiles are case-sensitive."
        )
    except Exception as e:
        raise RuntimeError(
            f"There was an error reading the Minikube configuration: {e}"
        )
    driver = config["Driver"]
    if driver == "docker":
        try:
            network_name = config["Network"] or config["Name"]
        except KeyError:
            network_name = "minikube"
    elif driver in ["kvm", "kvm2", "virtualbox"]:
        network_name = None
    else:
        raise RuntimeError(
            f"Gefyra does not support Minikube with this driver {driver}"
        )

    endpoint = _get_a_worker_ip(config)
    logger.debug(
        f"Minikube setup with driver '{driver}' network '{network_name}' and endpoint '{endpoint}'"
    )

    configuration_parameters = {
        "network_name": network_name,
        "cargo_endpoint_host": endpoint,
        "cargo_endpoint_port": "31820",
        "kube_context": config["Name"],
    }
    return configuration_parameters



import logging

from gefyra.configuration import default_configuration

from .utils import stopwatch


logger = logging.getLogger(__name__)


@stopwatch
def down(config=default_configuration) -> bool:
    from gefyra.cluster.manager import uninstall_operator
    from gefyra.local.cargo import remove_cargo_container
    from gefyra.local.networking import (
        handle_remove_network,
        kill_remainder_container_in_network,
    )
    from gefyra.local.bridge import remove_interceptrequest_remainder
    from gefyra.local.utils import (
        set_gefyra_network_from_cargo,
        set_kubeconfig_from_cargo,
    )

    try:
        config = set_kubeconfig_from_cargo(config)
        config = set_gefyra_network_from_cargo(config)
    except RuntimeError:
        logger.info("Gefyra client is not running.")
    try:
        logger.info("Removing running bridges")
        remove_interceptrequest_remainder(config)
        logger.info("Uninstalling Operator")
        uninstall_operator(config)
        logger.info("Removing Cargo")
    except Exception as e:
        logger.error(f"Could not remove all Gefyra cluster components: {e}")

    remove_cargo_container(config)
    kill_remainder_container_in_network(config, config.NETWORK_NAME)
    handle_remove_network(config)
    return True



import base64
import logging
from collections.abc import Mapping
from time import sleep

from gefyra.configuration import ClientConfiguration


logger = logging.getLogger(__name__)


def decode_secret(u):
    n = {}
    for k, v in u.items():
        if isinstance(v, Mapping):
            n[k] = decode_secret(v)
        else:
            n[k] = (base64.b64decode(v.encode("utf-8"))).decode("utf-8")
    return n


def get_env_from_pod_container(
    config: ClientConfiguration, pod_name: str, namespace: str, container_name: str
):
    from kubernetes.client import ApiException
    from kubernetes.stream import stream

    retries = 10
    counter = 0
    interval = 1
    while counter < retries:
        try:
            resp = stream(
                config.K8S_CORE_API.connect_get_namespaced_pod_exec,
                pod_name,
                namespace,
                container=container_name,
                command=["env"],
                stderr=True,
                stdin=False,
                stdout=True,
                tty=False,
            )
            return resp
        except ApiException as e:
            # e.status is 0 for some reason
            if "500 Internal Server Error" in e.reason:
                sleep(interval)
                counter += 1
                logger.debug(
                    f"Failed to get env from pod {pod_name} in namespace {namespace} on try {counter}."
                )
            else:
                raise e
    raise RuntimeError(
        f"Failed to get env from pod {pod_name} in namespace {namespace} after {retries} tries."
    )


def get_container(pod, container_name: str):
    for container in pod.spec.containers:
        if container.name == container_name:
            return container
    raise RuntimeError(f"Container {container_name} not found.")


def get_container_image(pod, container_name: str):
    container = get_container(pod, container_name=container_name)
    if container.image:
        return container.image
    raise RuntimeError(f"Container {container_name} image could not be determined.")


def get_container_command(pod, container_name: str):
    container = get_container(pod, container_name=container_name)
    res = []
    if container.command:
        res.extend(container.command)
    if container.args:
        res.extend(container.args)
    return " ".join(res)


def get_container_ports(pod, container_name: str):
    container = get_container(pod, container_name=container_name)
    if container.ports:
        return container.ports
    return []


def get_v1pod(
    config: ClientConfiguration,
    pod_name: str,
    namespace: str,
):
    from kubernetes.client import ApiException

    try:
        pod = config.K8S_CORE_API.read_namespaced_pod(pod_name, namespace)
    except ApiException as e:
        if e.status == 404:
            raise RuntimeError(
                f"Pod {pod_name} in namespace {namespace} does not exist."
            )
        else:
            raise e
    return pod



import logging
from dataclasses import dataclass
from enum import Enum

from gefyra.api import stopwatch
from gefyra.configuration import default_configuration, ClientConfiguration

logger = logging.getLogger(__name__)


@dataclass
class GefyraClusterStatus:
    # is a kubernetes cluster reachable
    connected: bool
    # is the operator running
    operator: bool
    operator_image: str
    # is stowaway running
    stowaway: bool
    stowaway_image: str
    # the gefyra namespace is available
    namespace: bool


@dataclass
class GefyraClientStatus:
    version: str
    # is cargo running
    cargo: bool
    cargo_image: str
    # is gefyra network available
    network: bool
    # is gefyra client connected with gefyra cluster
    connection: bool
    # amount of containers running in gefyra
    containers: int
    # amount of active bridges
    bridges: int
    # current kubeconfig file
    kubeconfig: str
    # current kubeconfig context
    context: str
    # wireguard endpoint
    cargo_endpoint: str


class StatusSummary(str, Enum):
    UP = "Gefyra is up and connected"
    DOWN = "Gefyra is not running"
    INCOMPLETE = "Gefyra is not running properly"


@dataclass
class GefyraStatus:
    summary: StatusSummary
    cluster: GefyraClusterStatus
    client: GefyraClientStatus


def _get_client_status(config: ClientConfiguration) -> GefyraClientStatus:
    from docker.errors import NotFound
    from gefyra.local import CARGO_ENDPOINT_LABEL, VERSION_LABEL

    # these are the default values
    _status = GefyraClientStatus(
        cargo=False,
        version="",
        cargo_image=config.CARGO_IMAGE,
        network=False,
        connection=False,
        containers=0,
        bridges=0,
        kubeconfig=config.KUBE_CONFIG_FILE,
        context=config.KUBE_CONTEXT,
        cargo_endpoint="",
    )
    try:
        logger.debug("Checking cargo container running")
        cargo_container = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
        if cargo_container.status == "running":
            _status.cargo = True
            _status.cargo_endpoint = cargo_container.attrs["Config"]["Labels"].get(
                CARGO_ENDPOINT_LABEL
            )
            _status.version = cargo_container.attrs["Config"]["Labels"].get(
                VERSION_LABEL
            )
            _status.cargo_image = cargo_container.image.tags[0]
    except NotFound:
        pass
    try:
        logger.debug("Checking gefyra network available")
        gefyra_net = config.DOCKER.networks.get(config.NETWORK_NAME)
        _status.network = True
        _status.containers = (
            len(gefyra_net.containers) - 1
            if _status.cargo is True
            else len(gefyra_net.containers)
        )
    except NotFound:
        return _status

    from gefyra.local.cargo import probe_wireguard_connection

    try:
        logger.debug("Probing wireguard connection")
        if _status.cargo:
            probe_wireguard_connection(config)
            _status.connection = True
    except RuntimeError:
        return _status

    from gefyra.local.bridge import get_all_interceptrequests

    logger.debug("Counting all active bridges")
    _status.bridges = len(get_all_interceptrequests(config))
    return _status


def _get_cluster_status(config: ClientConfiguration) -> GefyraClusterStatus:
    from kubernetes.client import ApiException
    from kubernetes.config import ConfigException
    from urllib3.exceptions import MaxRetryError

    # these are the default values
    _status = GefyraClusterStatus(
        connected=False,
        operator=False,
        operator_image="",
        stowaway=False,
        stowaway_image="",
        namespace=False,
    )
    # check if connected to the cluster
    try:
        logger.debug("Reading API resources from Kubernetes")
        config.K8S_CORE_API.get_api_resources(_request_timeout=(1, 5))
        _status.connected = True
    except (ApiException, ConfigException, MaxRetryError):
        return _status
    # check if gefyra namespace is available
    try:
        logger.debug("Reading gefyra namespace")
        config.K8S_CORE_API.read_namespace(
            name=config.NAMESPACE, _request_timeout=(1, 5)
        )
        _status.namespace = True
    except ApiException:
        return _status
    # check if the Gefyra operator is running and ready
    try:
        logger.debug("Checking operator deployment")
        operator_deploy = config.K8S_APP_API.read_namespaced_deployment(
            name="gefyra-operator", namespace=config.NAMESPACE, _request_timeout=(1, 5)
        )
        if (
            operator_deploy.status.ready_replicas
            and operator_deploy.status.ready_replicas >= 1
        ):
            _status.operator = True
            _status.operator_image = operator_deploy.spec.template.spec.containers[
                0
            ].image
    except ApiException:
        return _status

    # check if the Gefyra operator is running and ready
    try:
        logger.debug("Checking Stowaway deployment")
        stowaway_deploy = config.K8S_APP_API.read_namespaced_deployment(
            name="gefyra-stowaway", namespace=config.NAMESPACE, _request_timeout=(1, 5)
        )
        if (
            stowaway_deploy.status.ready_replicas
            and stowaway_deploy.status.ready_replicas >= 1
        ):
            _status.stowaway = True
            _status.stowaway_image = stowaway_deploy.spec.template.spec.containers[
                0
            ].image
    except ApiException:
        pass

    return _status


@stopwatch
def status(config=default_configuration) -> GefyraStatus:
    from gefyra.local.utils import set_kubeconfig_from_cargo

    # Check if kubeconfig is available through running Cargo
    config = set_kubeconfig_from_cargo(config)

    cluster = _get_cluster_status(config)
    client = _get_client_status(config)
    if client.connection:
        summary = StatusSummary.UP
    else:
        if client.cargo or (
            cluster.connected and cluster.operator and cluster.stowaway
        ):
            summary = StatusSummary.INCOMPLETE
        else:
            summary = StatusSummary.DOWN
    return GefyraStatus(cluster=cluster, client=client, summary=summary)



# flake8: noqa
import io
import platform
import sys


def get_dockerfile(cargo_image):
    run_patch = ""
    if sys.platform == "win32" or "microsoft" in platform.release().lower():
        run_patch = "RUN patch /usr/bin/wg-quick /wgquick.patch"
    return io.BytesIO(
        f""" 
FROM {cargo_image}
{run_patch}

ARG ADDRESS
ARG MTU
ARG PRIVATE_KEY
ARG DNS
ARG PUBLIC_KEY
ARG ENDPOINT
ARG ALLOWED_IPS

RUN echo '[Interface] \\n\
Address = '"$ADDRESS"' \\n\
MTU = '"$MTU"' \\n\
PrivateKey = '"$PRIVATE_KEY"' \\n\
DNS = '"$DNS"' \\n\
PreUp = sysctl -w net.ipv4.ip_forward=1 \\n\
PostUp = iptables -A FORWARD -i %i -j ACCEPT; iptables -A FORWARD -o %i -j ACCEPT; iptables -t nat -A POSTROUTING -o eth1 -j MASQUERADE \\n\ 
PostDown = iptables -D FORWARD -i %i -j ACCEPT; iptables -D FORWARD -o %i -j ACCEPT; iptables -t nat -D POSTROUTING -o eth1 -j MASQUERADE \\n\
\\n\
[Peer] \\n\
PublicKey = '"$PUBLIC_KEY"' \\n\
Endpoint = '"$ENDPOINT"' \\n\
PersistentKeepalive = 21 \\n\
AllowedIPs = '"$ALLOWED_IPS" > /config/wg0.conf

RUN cat /config/wg0.conf
""".encode(
            "utf-8"
        )
    )



from .bridge import *  # noqa
from .down import *  # noqa
from .up import *  # noqa
from .list import *  # noqa
from .reflect import *  # noqa
from .run import *  # noqa
from .status import *  # noqa



import logging
from typing import List, Dict, Union

from kubernetes.client import (
    V1ServiceAccount,
    V1ObjectMeta,
    V1ClusterRole,
    V1PolicyRule,
    V1ClusterRoleBinding,
    V1RoleRef,
    V1Subject,
    V1PodTemplateSpec,
    V1Deployment,
    V1PodSpec,
    V1Container,
    V1EnvVar,
    V1DeploymentSpec,
    ApiException,
    V1StatefulSet,
    V1Pod,
)

from gefyra.api.utils import get_workload_type
from gefyra.configuration import ClientConfiguration

logger = logging.getLogger(__name__)


def create_operator_serviceaccount(namespace: str) -> V1ServiceAccount:
    return V1ServiceAccount(
        metadata=V1ObjectMeta(
            # this name is referenced by Operator
            name="gefyra-operator",
            namespace=namespace,
        )
    )


def create_operator_clusterrole() -> V1ClusterRole:
    crd_rule = V1PolicyRule(
        api_groups=["apiextensions.k8s.io"],
        resources=["customresourcedefinitions"],
        verbs=["create", "patch", "delete", "list", "watch"],
    )
    kopf_rules = V1PolicyRule(
        api_groups=["kopf.dev"],
        resources=["clusterkopfpeerings"],
        verbs=["create", "patch"],
    )
    misc_res_rule = V1PolicyRule(
        api_groups=["", "apps", "extensions", "events.k8s.io"],
        resources=[
            "namespaces",
            "configmaps",
            "secrets",
            "deployments",
            "services",
            "serviceaccounts",
            "pods",
            "pods/exec",
            "events",
        ],
        verbs=["create", "patch", "update", "delete", "get", "list"],
    )
    ireq_rule = V1PolicyRule(
        api_groups=["gefyra.dev"], resources=["interceptrequests"], verbs=["*"]
    )

    clusterrole = V1ClusterRole(
        kind="ClusterRole",
        metadata=V1ObjectMeta(
            # this name is referenced by Operator
            name="gefyra-operator-role",
        ),
        rules=[
            crd_rule,
            kopf_rules,
            # kopf_peering,
            misc_res_rule,
            ireq_rule,
        ],
    )
    return clusterrole


def create_operator_clusterrolebinding(
    serviceaccount: V1ServiceAccount,
    clusterrole: V1ClusterRole,
    namespace: str,
) -> V1ClusterRoleBinding:
    return V1ClusterRoleBinding(
        metadata=V1ObjectMeta(
            # this name is referenced by Operator
            name="gefyra-operator-rolebinding",
        ),
        role_ref=V1RoleRef(
            api_group="rbac.authorization.k8s.io",
            name=clusterrole.metadata.name,
            kind="ClusterRole",
        ),
        subjects=[
            V1Subject(
                kind="ServiceAccount",
                name=serviceaccount.metadata.name,
                namespace=namespace,
            )
        ],
    )


def create_operator_deployment(
    serviceaccount: V1ServiceAccount,
    config: ClientConfiguration,
    gefyra_network_subnet: str,
) -> V1Deployment:
    template = V1PodTemplateSpec(
        metadata=V1ObjectMeta(labels={"app": "gefyra-operator"}),
        spec=V1PodSpec(
            containers=[
                V1Container(
                    name="gefyra-operator",
                    image=config.OPERATOR_IMAGE,
                    env=[
                        V1EnvVar(
                            name="GEFYRA_PEER_SUBNET", value=gefyra_network_subnet
                        ),
                        V1EnvVar(
                            name="GEFYRA_STOWAWAY_IMAGE",
                            value=config.STOWAWAY_IMAGE.split(":")[0],
                        ),
                        V1EnvVar(
                            name="GEFYRA_STOWAWAY_TAG",
                            value=config.STOWAWAY_IMAGE.split(":")[1],
                        ),
                        V1EnvVar(
                            name="GEFYRA_CARRIER_IMAGE",
                            value=config.CARRIER_IMAGE.split(":")[0],
                        ),
                        V1EnvVar(
                            name="GEFYRA_CARRIER_TAG",
                            value=config.CARRIER_IMAGE.split(":")[1],
                        ),
                    ],
                )
            ],
            service_account_name=serviceaccount.metadata.name,
        ),
    )
    spec = V1DeploymentSpec(
        replicas=1,
        template=template,
        selector={"matchLabels": {"app": "gefyra-operator"}},
    )
    deployment = V1Deployment(
        api_version="apps/v1",
        kind="Deployment",
        metadata=V1ObjectMeta(name="gefyra-operator", namespace=config.NAMESPACE),
        spec=spec,
    )

    return deployment


def _check_pod_for_command(pod: V1Pod, container_name: str):
    containers: list[V1Container] = pod.spec.containers
    if not len(containers):
        raise RuntimeError(f"No container available in pod {pod.metadata.name}.")

    ALLOWED_COMMANDS = [
        "sh",
        "bash",
        "zsh",
        "ash",
        "/bin/sh",
        "/bin/bash",
        "/bin/zsh",
        "/bin/ash",
        "/entrypoint.sh",
    ]
    for container in containers:
        if (
            container.name == container_name
            and container.command
            and container.command[0] not in ALLOWED_COMMANDS
        ):
            raise RuntimeError(
                f"Cannot bridge pod {pod.metadata.name} since it has a `command` defined."
            )


def check_pod_valid_for_bridge(
    config: ClientConfiguration, pod_name: str, namespace: str, container_name: str
):
    pod = config.K8S_CORE_API.read_namespaced_pod(
        name=pod_name,
        namespace=namespace,
    )

    _check_pod_for_command(pod, container_name)


def owner_reference_consistent(
    pod: V1Pod,
    workload: Union[V1Deployment, V1StatefulSet],
    config: ClientConfiguration,
) -> bool:
    if workload.kind == "StatefulSet":
        return pod.metadata.owner_references[0].uid == workload.metadata.uid
    elif workload.kind == "Deployment":
        try:
            replicaset_set = config.K8S_APP_API.read_namespaced_replica_set(
                name=pod.metadata.owner_references[0].name,
                namespace=pod.metadata.namespace,
            )
        except ApiException as e:
            if e.status == 404:
                return False
        return replicaset_set.metadata.owner_references[0].uid == workload.metadata.uid
    raise RuntimeError(
        f"Unknown workload type for owner reference check: {workload.kind}/{workload.metadata.name}."
    )


def get_pods_and_containers_for_workload(
    config: ClientConfiguration, name: str, namespace: str, workload_type: str
) -> Dict[str, List[str]]:
    workload = None
    result = {}
    API_EXCEPTION_MSG = "Exception when calling Kubernetes API: {}"
    NOT_FOUND_MSG = f"{workload_type.capitalize()} not found."
    workload_type = get_workload_type(workload_type)
    try:
        if workload_type == "deployment":
            workload = config.K8S_APP_API.read_namespaced_deployment(
                name=name, namespace=namespace
            )
        elif workload_type == "statefulset":
            workload = config.K8S_APP_API.read_namespaced_stateful_set(
                name=name, namespace=namespace
            )
    except ApiException as e:
        if e.status == 404:
            raise RuntimeError(NOT_FOUND_MSG)
        raise RuntimeError(API_EXCEPTION_MSG.format(e))

    # use workloads metadata uuid for owner references with field selector to get pods
    v1_label_selector = workload.spec.selector.match_labels

    label_selector = ",".join(
        [f"{key}={value}" for key, value in v1_label_selector.items()]
    )

    if not label_selector:
        raise RuntimeError(f"No label selector set for {workload_type} - {name}.")

    pods = config.K8S_CORE_API.list_namespaced_pod(
        namespace=namespace, label_selector=label_selector
    )
    for pod in pods.items:
        if owner_reference_consistent(pod, workload, config):
            result[pod.metadata.name] = [
                container.name for container in pod.spec.containers
            ]
    return result


def get_pods_and_containers_for_pod_name(
    config: ClientConfiguration, name: str, namespace: str
) -> Dict[str, List[str]]:
    result = {}
    API_EXCEPTION_MSG = "Exception when calling Kubernetes API: {}"
    try:
        pod = config.K8S_CORE_API.read_namespaced_pod(name=name, namespace=namespace)
    except ApiException as e:
        if e.status == 404:
            API_EXCEPTION_MSG = f"Pod {name} not found."
        raise RuntimeError(API_EXCEPTION_MSG.format(e))
    if not pod.spec:  # `.spec` is optional in python kubernetes client
        raise RuntimeError(f"Could not retrieve spec for pod - {name}.")
    result[pod.metadata.name] = [container.name for container in pod.spec.containers]
    return result



import logging

from gefyra.configuration import default_configuration, ClientConfiguration

logger = logging.getLogger("gefyra")


def probe_docker(config: ClientConfiguration = default_configuration):
    logger.info("Checking Docker client.")
    try:
        config.DOCKER.containers.list()
        logger.info("Docker client: Ok")
        logger.info("Checking availability of Gefyra Cargo image...")
        config.DOCKER.images.pull("quay.io/gefyra/cargo:latest")
        logger.info("Gefyra Cargo: Available")
    except Exception:
        logger.error("Docker does not seem to be not working for Gefyra")
        return False
    else:
        logger.info("Docker: Ok")
        return True


def probe_kubernetes(config: ClientConfiguration = default_configuration):
    logger.info("Checking Kubernetes connection.")
    try:
        config.K8S_CORE_API.list_namespace()
    except Exception:
        logger.error(
            "Kubernetes is not connected to a cluster or does not seem to be working for Gefyra"
        )
        return False
    else:
        logger.info("Kubernetes: Ok")
        return True



import logging
import socket
import time
from typing import Iterable

logger = logging.getLogger(__name__)


def is_port_free(port):
    """
    Check if a port is free on the current system.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("127.0.0.1", port))
            return True
        except OSError:
            return False


def get_workload_type(workload_type_str: str):
    POD = ["pod", "po", "pods"]
    DEPLOYMENT = ["deploy", "deployment", "deployments"]
    STATEFULSET = ["statefulset", "sts", "statefulsets"]
    VALID_TYPES = POD + DEPLOYMENT + STATEFULSET

    if workload_type_str not in VALID_TYPES:
        raise RuntimeError(
            f"Unknown workload type {workload_type_str}\n"
            f'Valid workload types include: {", ".join(str(valid_type) for valid_type in VALID_TYPES)}'
        )

    if workload_type_str in POD:
        return "pod"
    elif workload_type_str in DEPLOYMENT:
        return "deployment"
    elif workload_type_str in STATEFULSET:
        return "statefulset"


def generate_env_dict_from_strings(env_vars: Iterable[str]) -> dict:
    return {k[0]: k[1] for k in [arg.split("=", 1) for arg in env_vars] if len(k) > 1}


def stopwatch(func):
    def wrapper(*args, **kwargs):
        tic = time.perf_counter()
        result = func(*args, **kwargs)
        toc = time.perf_counter()
        logger.debug(
            f"Operation time for '{func.__name__}(...)' was {(toc - tic)*1000:0.4f}ms"
        )
        return result

    return wrapper



import logging
import time

from kubernetes.client import (
    V1ServiceAccount,
    V1ClusterRole,
    ApiException,
    V1ClusterRoleBinding,
    V1Deployment,
    V1Namespace,
    V1ObjectMeta,
)
from kubernetes.watch import Watch

from gefyra.configuration import ClientConfiguration

from .resources import (
    create_operator_clusterrole,
    create_operator_clusterrolebinding,
    create_operator_deployment,
    create_operator_serviceaccount,
)
from .utils import decode_secret

logger = logging.getLogger(__name__)


def _handle_duplicate_namespace(config: ClientConfiguration):
    active = False
    try:
        namespace = config.K8S_CORE_API.read_namespace(config.NAMESPACE)
    except ApiException as e:
        if e.status == 404:
            logger.warning(f"NS {config.NAMESPACE} after 409. Retrying.")
        else:
            raise e
    if namespace.status.phase == "Active":
        active = True
    return active


def handle_create_namespace(config: ClientConfiguration, retries=10, wait=3):
    counter = 0
    created = False
    while counter < retries:
        try:
            config.K8S_CORE_API.create_namespace(
                body=V1Namespace(
                    metadata=V1ObjectMeta(
                        name=config.NAMESPACE,
                        labels={
                            "pod-security.kubernetes.io/enforce": "privileged",
                        },
                    )
                )
            )
            created = True
            break
        except ApiException as e:
            if e.status == 409:
                active = _handle_duplicate_namespace(config)
                if active:
                    created = True
                    break
            else:
                raise e
        logger.warning(f"Could not create namespace {config.NAMESPACE}. Retrying.")
        counter += 1
        time.sleep(wait)
    if not created:
        raise RuntimeError(
            f"Could not create namespace. Retried {retries} times. API returned HTTP 409."
        )


def handle_serviceaccount(
    config: ClientConfiguration, serviceaccount: V1ServiceAccount
):
    retries = 5
    wait = 5
    counter = 0
    while counter < retries:
        try:
            config.K8S_CORE_API.create_namespaced_service_account(
                body=serviceaccount, namespace=config.NAMESPACE
            )
            break
        except ApiException as e:
            if e.status == 409:
                break
            elif e.status == 403:
                # this sometimes happens when to cluster in terminating the gefyra namespace
                # due to a previous `gefyra down`. As long as it is terminating this error occurs.
                if counter > retries:
                    raise RuntimeError(
                        f"You're not allowed to create a serviceaccount in namespace {config.NAMESPACE}."
                    )
                else:
                    counter += 1
                    time.sleep(wait)
                    continue
            else:
                raise e


def handle_clusterrole(config: ClientConfiguration, clusterrole: V1ClusterRole):
    try:
        config.K8S_RBAC_API.create_cluster_role(body=clusterrole)
    except ApiException as e:
        if e.status == 409:
            pass
        else:
            raise e


def handle_clusterrolebinding(
    config: ClientConfiguration, clusterrolebinding: V1ClusterRoleBinding
):
    try:
        config.K8S_RBAC_API.create_cluster_role_binding(body=clusterrolebinding)
    except ApiException as e:
        if e.status == 409:
            pass
        else:
            raise e


def handle_deployment(
    config: ClientConfiguration, operator_deployment: V1Deployment
) -> bool:
    try:
        config.K8S_APP_API.create_namespaced_deployment(
            body=operator_deployment, namespace=config.NAMESPACE
        )
        return True
    except ApiException as e:
        if e.status == 409:
            return False
        else:
            raise e


def install_operator(config: ClientConfiguration, gefyra_network_subnet: str) -> dict:
    """
    Installs Gefyra Operator to the configured cluster, waits for the installation to complete and returns the
    connection secrets for Cargo
    :param config: a ClientConfiguration install
    :return: Cargo connection details
    """
    tic = time.perf_counter()
    handle_create_namespace(config=config)

    serviceaccount = create_operator_serviceaccount(config.NAMESPACE)
    clusterrole = create_operator_clusterrole()
    clusterrolebinding = create_operator_clusterrolebinding(
        serviceaccount, clusterrole, config.NAMESPACE
    )
    operator_deployment = create_operator_deployment(
        serviceaccount, config, f"{gefyra_network_subnet}"
    )
    handle_serviceaccount(config, serviceaccount)
    handle_clusterrole(config, clusterrole)
    handle_clusterrolebinding(config, clusterrolebinding)
    created = handle_deployment(config, operator_deployment)

    if created:
        w = Watch()

        # block (forever) until Gefyra cluster side is ready
        for event in w.stream(
            config.K8S_CORE_API.list_namespaced_event, namespace=config.NAMESPACE
        ):
            if event["object"].reason in ["Pulling", "Pulled"]:
                logger.info(event["object"].message)
            if event["object"].reason == "Gefyra-Ready":
                toc = time.perf_counter()
                logger.info(f"Operator became ready in {toc - tic:0.4f} seconds")
                break
    else:
        logger.info("Gefyra Operator already exists")

    cargo_connection_secret = config.K8S_CORE_API.read_namespaced_secret(
        name="gefyra-cargo-connection", namespace=config.NAMESPACE
    )
    values = decode_secret(cargo_connection_secret.data)
    logger.debug("Cargo connection details")
    logger.debug(values)
    return values


def uninstall_operator(config: ClientConfiguration):
    serviceaccount = create_operator_serviceaccount(config.NAMESPACE)
    clusterrole = create_operator_clusterrole()
    clusterrolebinding = create_operator_clusterrolebinding(
        serviceaccount, clusterrole, config.NAMESPACE
    )
    operator_deployment = create_operator_deployment(serviceaccount, config, "")
    try:
        config.K8S_APP_API.delete_namespaced_deployment(
            name=operator_deployment.metadata.name, namespace=config.NAMESPACE
        )
        # pause to let Operator shutdown properly
        time.sleep(5)
        config.K8S_CORE_API.delete_namespaced_service_account(
            name=serviceaccount.metadata.name, namespace=config.NAMESPACE
        )
        config.K8S_RBAC_API.delete_cluster_role(name=clusterrole.metadata.name)
        config.K8S_RBAC_API.delete_cluster_role_binding(
            name=clusterrolebinding.metadata.name
        )
    except ApiException:
        pass

    try:
        config.K8S_CORE_API.delete_namespace(name=config.NAMESPACE)
    except ApiException as e:
        if e.status == 404:
            pass
        else:
            raise e



import logging

from docker.errors import NotFound, APIError
from docker.models.networks import Network
from docker.types import IPAMConfig, IPAMPool

from gefyra.configuration import ClientConfiguration
from gefyra.local import CREATED_BY_LABEL

logger = logging.getLogger(__name__)


def create_gefyra_network(config: ClientConfiguration) -> Network:
    gefyra_network = handle_create_network(config)
    logger.debug(f"Network {gefyra_network.attrs}")
    return gefyra_network


def handle_create_network(config: ClientConfiguration) -> Network:
    DOCKER_MTU_OPTION = "com.docker.network.driver.mtu"
    try:
        network = config.DOCKER.networks.get(config.NETWORK_NAME)
        logger.info("Gefyra network already exists")
        if (
            CREATED_BY_LABEL[0] not in network.attrs["Labels"]
            or network.attrs["Labels"][CREATED_BY_LABEL[0]] != "true"
        ):
            logger.debug(f"Docker network '{network.name}' is not managed by Gefyra")
        if (
            "Options" in network.attrs
            and DOCKER_MTU_OPTION in network.attrs["Options"]
            and network.attrs["Options"][DOCKER_MTU_OPTION] != config.WIREGUARD_MTU
        ) or (
            "Options" in network.attrs
            and DOCKER_MTU_OPTION not in network.attrs["Options"]
        ):
            _mtu = (
                network.attrs["Options"].get(DOCKER_MTU_OPTION)
                if "Options" in network.attrs
                else "default"
            )
            logger.warning(
                f"The MTU value of the 'gefyra' network (={_mtu}) is different from the --wireguard-mtu parameter "
                f"(={config.WIREGUARD_MTU}) or default. You may experience bad network connections. Consider removing "
                f"the network 'gefyra' with 'docker network rm gefyra' before running 'gefyra up'."
            )
        return network
    except NotFound:
        pass

    # this is a workaround to select a free subnet (instead of finding it with python code)
    temp_network = config.DOCKER.networks.create(config.NETWORK_NAME, driver="bridge")
    subnet = temp_network.attrs["IPAM"]["Config"][0]["Subnet"]
    temp_network.remove()  # remove the temp network again

    ipam_pool = IPAMPool(subnet=f"{subnet}", aux_addresses={})
    ipam_config = IPAMConfig(pool_configs=[ipam_pool])
    network = config.DOCKER.networks.create(
        config.NETWORK_NAME,
        driver="bridge",
        ipam=ipam_config,
        labels={
            CREATED_BY_LABEL[0]: CREATED_BY_LABEL[1],
        },
        options={DOCKER_MTU_OPTION: config.WIREGUARD_MTU},
    )
    logger.info(f"Created network '{config.NETWORK_NAME}' ({network.short_id})")
    return network


def handle_remove_network(config: ClientConfiguration) -> None:
    """Removes all docker networks with the given name."""
    # we would need the id to identify the network unambiguously, so we just remove all networks that can be found with
    # the given name, under the assumption that no other docker network inadvertently uses the same name
    try:
        gefyra_network = config.DOCKER.networks.get(config.NETWORK_NAME)
        if (
            CREATED_BY_LABEL[0] in gefyra_network.attrs["Labels"]
            and gefyra_network.attrs["Labels"][CREATED_BY_LABEL[0]] == "true"
        ):
            logger.info(f"Removing Docker network {gefyra_network.name}")
            gefyra_network.remove()
        else:
            logger.info(
                f"Docker network {gefyra_network.name} is not managed by Gefyra"
            )
    except NotFound:
        pass
    except APIError as e:
        logger.error(f"Could not remove network due to the following error: {e}")


def kill_remainder_container_in_network(
    config: ClientConfiguration, network_name
) -> None:
    """Kills all containers from this network"""
    try:
        network = config.DOCKER.networks.get(network_name)
        containers = network.attrs["Containers"].keys()
        for container in containers:
            c = config.DOCKER.containers.get(container)
            if (
                CREATED_BY_LABEL[0] in c.attrs["Config"]["Labels"]
                and c.attrs["Config"]["Labels"][CREATED_BY_LABEL[0]] == "true"
            ):
                c.kill()
    except NotFound:
        pass



import logging
import os
import sys
from threading import Thread

from gefyra.configuration import default_configuration, ClientConfiguration
from .utils import generate_env_dict_from_strings, stopwatch, get_workload_type


logger = logging.getLogger(__name__)


def pod_ready_and_healthy(
    config: ClientConfiguration, pod_name: str, namespace: str, container_name: str
):
    pod = config.K8S_CORE_API.read_namespaced_pod_status(pod_name, namespace=namespace)

    container_idx = next(
        i
        for i, container_status in enumerate(pod.status.container_statuses)
        if container_status.name == container_name
    )

    return (
        pod.status.phase == "Running"
        and pod.status.container_statuses[container_idx].ready
        and pod.status.container_statuses[container_idx].started
        and pod.status.container_statuses[container_idx].state.running
        and pod.status.container_statuses[container_idx].state.running.started_at
    )


def retrieve_pod_and_container(
    env_from: str, namespace: str, config: ClientConfiguration
) -> (str, str):
    from gefyra.cluster.resources import (
        get_pods_and_containers_for_workload,
        get_pods_and_containers_for_pod_name,
    )

    container_name = ""
    workload_type, workload_name = env_from.split("/", 1)

    workload_type = get_workload_type(workload_type)

    if "/" in workload_name:
        workload_name, container_name = workload_name.split("/")

    if workload_type != "pod":
        pods = get_pods_and_containers_for_workload(
            config, name=workload_name, namespace=namespace, workload_type=workload_type
        )
    else:
        pods = get_pods_and_containers_for_pod_name(
            config=config, name=workload_name, namespace=namespace
        )

    while len(pods):
        pod_name, containers = pods.popitem()
        if container_name and container_name not in containers:
            raise RuntimeError(
                f"{container_name} was not found for {workload_type}/{workload_name}"
            )
        actual_container_name = container_name or containers[0]
        if pod_ready_and_healthy(config, pod_name, namespace, actual_container_name):
            return pod_name, actual_container_name

    raise RuntimeError(
        f"Could not find a ready pod for {workload_type}/{workload_name}"
    )


def print_logs(container):
    for logline in container.logs(stream=True):
        print(logline.decode("utf-8"), end="")


@stopwatch
def run(
    image: str,
    name: str = None,
    command: str = None,
    volumes: dict = None,
    ports: dict = None,
    detach: bool = True,
    auto_remove: bool = False,
    namespace: str = None,
    env: list = None,
    env_from: str = None,
    config=default_configuration,
) -> bool:
    from kubernetes.client import ApiException
    from docker.errors import APIError
    from gefyra.cluster.utils import get_env_from_pod_container
    from gefyra.local.bridge import deploy_app_container
    from gefyra.local.utils import (
        get_processed_paths,
        set_gefyra_network_from_cargo,
        set_kubeconfig_from_cargo,
    )
    from gefyra.local.cargo import probe_wireguard_connection

    # Check if kubeconfig is available through running Cargo
    config = set_kubeconfig_from_cargo(config)

    # #125: Fallback to namespace in kube config
    if namespace is None:
        from kubernetes.config import kube_config

        _, active_context = kube_config.list_kube_config_contexts()
        namespace = active_context["context"].get("namespace") or "default"
        ns_source = "kubeconfig"
    else:
        ns_source = "--namespace argument"

    dns_search = f"{namespace}.svc.cluster.local"
    config = set_gefyra_network_from_cargo(config)
    #
    # Confirm the wireguard connection working
    #
    try:
        probe_wireguard_connection(config)
    except Exception as e:
        logger.error(e)
        return False

    volumes = get_processed_paths(os.getcwd(), volumes)
    #
    # 1. get the ENV together a) from a K8s container b) from override
    #
    env_dict = {}
    try:
        if env_from:
            env_from_pod, env_from_container = retrieve_pod_and_container(
                env_from, namespace=namespace, config=config
            )
            logger.debug(f"Using ENV from {env_from_pod}/{env_from_container}")
            raw_env = get_env_from_pod_container(
                config, env_from_pod, namespace, env_from_container
            )
            logger.debug("ENV from pod/container is:\n" + raw_env)
            raw_env_vars = raw_env.split("\n")
            env_dict = generate_env_dict_from_strings(raw_env_vars)
    except ApiException as e:
        logger.error(f"Cannot copy environment from Pod: {e.reason} ({e.status}).")
        return False
    if env:
        env_overrides = generate_env_dict_from_strings(env)
        env_dict.update(env_overrides)

    #
    # 2. deploy the requested container to Gefyra
    #
    try:
        container = deploy_app_container(
            config,
            image,
            name,
            command,
            volumes,
            ports,
            env_dict,
            auto_remove,
            dns_search,
        )
    except APIError as e:
        if e.status_code == 409:
            logger.warning(e.explanation)
            return True
        else:
            raise RuntimeError(e.explanation)

    logger.info(
        f"Container image '{', '.join(container.image.tags)}' started with name '{container.name}' in "
        f"Kubernetes namespace '{namespace}' (from {ns_source})"
    )
    if detach:
        return True
    else:
        try:
            logger.debug("Now printing out logs")
            t = Thread(target=print_logs, args=[container], daemon=True)
            t.start()
            while True:
                line = sys.stdin.readline()
                if not line:
                    logger.info(f"Detached from container: {name}")
                    return True
        except KeyboardInterrupt:
            container.stop()
            logger.info(f"Container stopped: {name}")



import logging
import configparser
import uuid
from pathlib import Path

from cli_tracker.sdk import CliTracker
from gefyra.configuration import __VERSION__

logger = logging.getLogger("gefyra")

#########################################################################################
# Telemetry information
# We are collecting anonymous data about the usage of Gefyra as CLI
# All the data we collect is published here: ``
# This data is supposed to help us develop Gefyra and make it a better tool, prioritize
# issues and work on bugs, features and review of pull requests.
# If you do not which to send telemetry data this is totally fine.
# Just opt out via `gefyra telemetry --off`.
#########################################################################################

SENTRY_DSN = (
    "https://97c8c0409cb74a079a93f05021b329f0@o146863.ingest.sentry.io/4505119985172480"
)


class CliTelemetry:
    dir_name = ".gefyra"
    file_name = "config.ini"

    def __init__(self):
        # We're loading / creating a settings file in the home directory.
        home = Path.home()
        gefyra_dir = home / self.dir_name
        if gefyra_dir.exists():
            gefyra_settings_path = gefyra_dir / self.file_name
            if gefyra_settings_path.exists():
                config = self.load_config(str(gefyra_settings_path))
            else:
                config = self.create_config(gefyra_settings_path)
        else:
            config = self.create_config(gefyra_dir / self.file_name)
        try:
            config["telemetry"].getboolean("track")
            # This was added later and is here for backwards compatbility
            user_id = self._get_user_id(config)
        except KeyError:
            config = self.create_config(gefyra_dir / self.file_name)

        if config["telemetry"].getboolean("track"):
            self._init_tracker(user_id=user_id)

    def _init_tracker(self, user_id):
        self.tracker = CliTracker(
            application="gefyra",
            dsn=SENTRY_DSN,
            release=__VERSION__,
            fingerprint=user_id,
        )

    def _get_user_id(self, config):
        if config["telemetry"].get("id"):
            user_id = config["telemetry"].get("id")
        else:
            user_id = self.create_id()
        return user_id

    def load_config(self, path):
        config = configparser.ConfigParser()
        config.read(path)
        self.path = path
        return config

    def create_config(self, path):
        config = configparser.ConfigParser()
        config["telemetry"] = {"track": "True", "id": str(uuid.uuid4())}
        output_file = Path(path)
        output_file.parent.mkdir(exist_ok=True, parents=True)

        with open(str(output_file), "w") as config_file:
            config.write(config_file)
        self.path = path
        return config

    def create_id(self):
        config = configparser.ConfigParser()
        config.read(self.path)
        user_id = str(uuid.uuid4())
        config["telemetry"]["id"] = user_id
        with open(str(self.path), "w") as config_file:
            config.write(config_file)
        return user_id

    def off(self):
        config = configparser.ConfigParser()
        config.read(self.path)
        config["telemetry"]["track"] = "False"
        with open(str(self.path), "w") as config_file:
            config.write(config_file)
        if hasattr(self, "tracker"):
            self.tracker.report_opt_out()
        logger.info("Disabled telemetry.")

    def on(self, test=False):
        config = configparser.ConfigParser()
        config.read(self.path)
        config["telemetry"]["track"] = "True"
        with open(str(self.path), "w") as config_file:
            config.write(config_file)
        user_id = self._get_user_id(config)
        self._init_tracker(user_id=user_id)
        if not test:
            self.tracker.report_opt_in()
        logger.info("Enabled telemetry.")



#!/usr/bin/env python3
import argparse
import logging
import traceback

from gefyra.api import get_containers_and_print, get_bridges_and_print, GefyraStatus
from gefyra.configuration import ClientConfiguration
from gefyra.local.utils import (
    PortMappingParser,
    IpPortMappingParser,
    get_connection_from_kubeconfig,
)
from gefyra.local.minikube import detect_minikube_config
from gefyra.local.telemetry import CliTelemetry

logger = logging.getLogger("gefyra")
parser = argparse.ArgumentParser(
    prog="gefyra",
    description="The Gefyra client. For more help please visit: https://gefyra.dev/reference/cli/",
)
action = parser.add_subparsers(dest="action", help="the action to be performed")
parser.add_argument("-d", "--debug", action="store_true", help="add debug output")

port_mapping_help_text = "Add port mapping in form of <container_port>:<host_port>"
namespace_help_text = "The namespace for this container to run in"

up_parser = action.add_parser("up")
up_parser.add_argument(
    "-e",
    "--endpoint",
    dest="cargo_endpoint",
    help="the Wireguard endpoint in the form <IP>:<Port> for Gefyra to connect to",
    required=False,
)
up_parser.add_argument(
    "-H",
    "--host",
    dest="cargo_endpoint_host",
    help="Hostname or IP of a K8s node for Gefyra to connect to."
    "Gefyra tries to extract this from the current kubeconfig and context.",
    required=False,
)
up_parser.add_argument(
    "-p",
    "--port",
    dest="cargo_endpoint_port",
    help="Open UDP port of the K8S node to connect to. Default to 31820.",
    required=False,
    default="31820",
)
up_parser.add_argument(
    "-M",
    "--minikube",
    help="Let Gefyra automatically find out the connection parameters for a "
    "local Minikube cluster for the given profile (default 'minikube').",
    required=False,
    nargs="?",
    const="minikube",
)
up_parser.add_argument(
    "-o",
    "--operator",
    dest="operator_image_url",
    help="Registry url for the operator image.",
    required=False,
)
up_parser.add_argument(
    "-s",
    "--stowaway",
    dest="stowaway_image_url",
    help="Registry url for the stowaway image.",
    required=False,
)
up_parser.add_argument(
    "-c",
    "--carrier",
    dest="carrier_image_url",
    help="Registry url for the carrier image.",
    required=False,
)
up_parser.add_argument(
    "-a",
    "--cargo",
    dest="cargo_image_url",
    help="Registry url for the cargo image.",
    required=False,
)
up_parser.add_argument(
    "-r",
    "--registry",
    dest="registry_url",
    help="Base url for registry to pull images from.",
    required=False,
)
up_parser.add_argument(
    "--wireguard-mtu",
    dest="wireguard_mtu",
    help="The MTU value for the local Wireguard endpoint (default: 1340).",
)
up_parser.add_argument(
    "--kubeconfig",
    dest="kube_config_file",
    required=False,
    help="The path to kubeconfig file",
)
up_parser.add_argument(
    "--context",
    dest="kube_context",
    required=False,
    help="The context name from kubeconfig",
)

reflect_parser = action.add_parser("reflect")
reflect_parser.add_argument(
    "-w", "--workload", help="Workload to reflect locally", required=True
)
reflect_parser.add_argument(
    "-e",
    "--expose",
    help="Exposes ports from k8s container spec to host",
    required=False,
    default=True,
    action="store_false",
)
reflect_parser.add_argument(
    "-i", "--image", help="The docker image to run in Gefyra", required=True
)
reflect_parser.add_argument(
    "-v",
    "--volume",
    action="append",
    help="Bind mount a volume into the container in notation src:dest, allowed multiple times",
    required=False,
)
reflect_parser.add_argument(
    "-p",
    "--port",
    help=port_mapping_help_text,
    required=False,
    action=PortMappingParser,
)
reflect_parser.add_argument(
    "-b",
    "--bridge",
    help="Bridge workload and make local container accessible to k8s workloads",
    required=False,
    default=False,
    action="store_true",
)

reflect_parser.add_argument(
    "-c",
    "--command",
    help="The command for this container to in Gefyra",
    nargs="+",
    required=False,
)
reflect_parser.add_argument(
    "-n",
    "--namespace",
    help=namespace_help_text,
    default="default",
)
reflect_parser.add_argument(
    "--env",
    action="append",
    help="Set or override environment variables in the form ENV=value, allowed multiple times",
    required=False,
)

run_parser = action.add_parser("run")
run_parser.add_argument(
    "-i", "--image", help="The docker image to run in Gefyra", required=True
)
run_parser.add_argument(
    "-N", "--name", help="The name of the container running in Gefyra", required=True
)
run_parser.add_argument(
    "-c",
    "--command",
    help="The command for this container to in Gefyra",
    nargs="+",
    required=False,
)
run_parser.add_argument(
    "-n",
    "--namespace",
    help=namespace_help_text,
)
run_parser.add_argument(
    "--env",
    action="append",
    help="Set or override environment variables in the form ENV=value, allowed multiple times",
    required=False,
)
run_parser.add_argument(
    "-v",
    "--volume",
    action="append",
    help="Bind mount a volume into the container in notation src:dest, allowed multiple times",
    required=False,
)
run_parser.add_argument(
    "--env-from",
    help="Copy the environment from the container in the notation 'Pod/Container'",
    required=False,
)
run_parser.add_argument(
    "-p",
    "--expose",
    help=port_mapping_help_text,
    required=False,
    action=IpPortMappingParser,
)
run_parser.add_argument(
    "--rm",
    help="Automatically remove the container when it exits",
    dest="auto_remove",
    action="store_true",
    default=False,
    required=False,
)
run_parser.add_argument(
    "-d",
    "--detach",
    help="Run container in background and print container ID",
    action="store_true",
    default=False,
    required=False,
)
bridge_parser = action.add_parser("bridge")
bridge_parser.add_argument(
    "-N", "--name", help="The name of the container running in Gefyra", required=True
)
bridge_parser.add_argument(
    "-p",
    "--port",
    help=port_mapping_help_text,
    required=True,
    action=PortMappingParser,
)
bridge_parser.add_argument(
    "-n",
    "--namespace",
    help=namespace_help_text,
    default="default",
)
bridge_parser.add_argument(
    "-P",
    "--no-probe-handling",
    action="store_true",
    help="Make Carrier to not handle probes during switch operation",
    default=False,
)
bridge_parser.add_argument(
    "--target",
    help="Intercept the container given in the notation 'resource/name/container'. "
    "Resource can be one of 'deployment', 'statefulset' or 'pod'. "
    "E.g.: --target deployment/hello-nginx/nginx",
    required=False,
)

unbridge_parser = action.add_parser("unbridge")
unbridge_parser.add_argument("-N", "--name", help="The name of the bridge")
unbridge_parser.add_argument(
    "-w",
    "--wait",
    help="Block until deletion is complete",
    action="store_true",
    default=False,
    required=False,
)
unbridge_parser.add_argument(
    "-A", "--all", help="Removes all active bridges", action="store_true"
)
list_parser = action.add_parser("list")
list_parser.add_argument(
    "--containers", help="List all containers running in Gefyra", action="store_true"
)
list_parser.add_argument(
    "--bridges", help="List all active bridges in Gefyra", action="store_true"
)
down_parser = action.add_parser("down")
check_parser = action.add_parser("check")
status_parser = action.add_parser("status")
version_parser = action.add_parser("version")

version_parser.add_argument(
    "-n",
    "--no-check",
    help="Do not check whether there is a new version",
    action="store_true",
    default=False,
)

telemetry_parser = action.add_parser("telemetry")
telemetry_parser.add_argument("--off", help="Turn off telemetry", action="store_true")
telemetry_parser.add_argument("--on", help="Turn on telemetry", action="store_true")

try:
    telemetry = CliTelemetry()
except Exception:  # pragma: no cover
    telemetry = False


def version(config, check: bool):
    import requests

    logger.info(f"Gefyra client version: {config.__VERSION__}")
    if check:
        release = requests.get(
            "https://api.github.com/repos/gefyrahq/gefyra/releases/latest"
        )
        if release.status_code == 403:  # pragma: no cover
            logger.info("Versions cannot be compared, as API rate limit was exceeded")
            return None
        latest_release_version = release.json()["tag_name"].replace("-", ".")
        if config.__VERSION__ != latest_release_version:  # pragma: no cover
            logger.info(
                f"You are using gefyra version {config.__VERSION__}; however, version {latest_release_version} is "
                f"available."
            )
    return True


def telemetry_command(on, off):
    if not telemetry:
        logger.info("Telemetry in not working on your machine. No action taken.")
        return
    if off and not on:
        telemetry.off()
    elif on and not off:
        telemetry.on()
    else:
        logger.info("Invalid flags. Please use either --off or --on.")


def get_client_configuration(args) -> ClientConfiguration:
    configuration_params = {}

    if args.action == "up":
        if args.cargo_endpoint:
            logger.warning(
                "`--endpoint`/`-e` has been removed. Please consider `--host` and `--port` instead."
            )
            exit(1)
        if args.minikube and bool(args.cargo_endpoint_host):
            raise RuntimeError("You cannot use --host together with --minikube.")

        if args.minikube:
            configuration_params.update(detect_minikube_config(args.minikube))
        else:
            if not args.cargo_endpoint_host:
                # #138: Read in the endpoint from kubeconfig
                endpoint = get_connection_from_kubeconfig(args.kube_config_file)
                if endpoint:
                    logger.info(f"Setting host and port from kubeconfig {endpoint}")
                    configuration_params["cargo_endpoint_host"] = endpoint.split(":")[0]
                    configuration_params["cargo_endpoint_port"] = endpoint.split(":")[1]
                else:
                    logger.info(
                        "There was no --host argument provided. Connecting to a local Kubernetes node."
                    )
        for argument in vars(args):
            if argument not in [
                "action",
                "debug",
                "minikube",
                "cargo_endpoint",
            ]:
                # don't overwrite an option which has been determined already
                if argument not in configuration_params:
                    logger.debug(
                        f"Setting remainder option: {argument}:{getattr(args, argument)}"
                    )
                    configuration_params[argument] = getattr(args, argument)

    configuration = ClientConfiguration(**configuration_params)
    logger.debug("ClientConfiguration: " + str(vars(configuration)))

    return configuration


def print_status(status: GefyraStatus):
    import json
    import dataclasses

    class EnhancedJSONEncoder(json.JSONEncoder):
        def default(self, o):
            # we only accept dataclasses here
            return dataclasses.asdict(o)

    print(json.dumps(status, cls=EnhancedJSONEncoder, indent=2))


def cli_up(configuration):
    from gefyra.api import up

    success = up(config=configuration)
    if not success:
        raise RuntimeError("Failed to start Gefyra")


def setup_logger(args, configuration_package):
    if args.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    logger.addHandler(configuration_package.console)


def main():
    try:
        from gefyra import configuration as configuration_package
        from gefyra.api import (
            bridge,
            down,
            run,
            unbridge,
            unbridge_all,
            status,
            reflect,
        )
        from gefyra.local.check import probe_kubernetes, probe_docker

        args = parser.parse_args()

        setup_logger(args, configuration_package)

        configuration = get_client_configuration(args)

        if args.action == "version":
            check = not args.no_check
            version(configuration_package, check)
            exit(0)
        elif args.action == "up":
            cli_up(configuration=configuration)
        elif args.action == "run":
            run(
                image=args.image,
                name=args.name,
                command=" ".join(args.command) if args.command else None,
                namespace=args.namespace,
                env_from=args.env_from,
                env=args.env,
                ports=args.expose,
                auto_remove=args.auto_remove,
                volumes=args.volume,
                config=configuration,
                detach=args.detach,
            )
        elif args.action == "reflect":
            reflect(
                workload=args.workload,
                expose_ports=args.expose,
                volumes=args.volume,
                command=args.command,
                namespace=args.namespace,
                do_bridge=args.bridge,
                ports=args.port,
                image=args.image,
            )
        elif args.action == "bridge":
            bridge(
                args.name,
                args.port,
                namespace=args.namespace,
                handle_probes=not args.no_probe_handling,
                config=configuration,
                target=args.target,
            )
        elif args.action == "unbridge":
            if args.name:
                unbridge(args.name, wait=args.wait, config=configuration)
            elif args.all:
                unbridge_all(wait=args.wait, config=configuration)
            else:
                logger.warning(
                    "Unbridge failed. Please use command with either -N or -A flag."
                )
                exit(1)
        elif args.action == "list":
            if args.containers:
                get_containers_and_print(config=configuration)
            elif args.bridges:
                get_bridges_and_print(config=configuration)
            else:
                get_containers_and_print(config=configuration)
                get_bridges_and_print(config=configuration)
        elif args.action == "down":
            down(config=configuration)
        elif args.action == "status":
            _status = status(config=configuration)
            print_status(_status)
        elif args.action == "check":
            probe_docker()
            probe_kubernetes(config=configuration)
        elif args.action == "telemetry":
            telemetry_command(on=args.on, off=args.off)
        else:
            parser.print_help()
    except KeyboardInterrupt:
        logger.warning("Program interrupted by user. Exiting...")
        exit(1)
    except Exception as e:
        if args.debug:
            traceback.print_exc()
        logger.fatal(f"There was an error running Gefyra: {e}")
        exit(1)
    exit(0)


if __name__ == "__main__":  # noqa
    main()



import logging
from tabulate import tabulate
from typing import List

from gefyra.configuration import default_configuration

from .utils import stopwatch


logger = logging.getLogger(__name__)


def get_containers_and_print(config: default_configuration):
    from gefyra.api import list_containers

    containers = list_containers(config=config)
    print(
        tabulate(
            containers, headers=["NAME", "IP ADDRESS", "NAMESPACE"], tablefmt="plain"
        )
    )


def get_bridges_and_print(config: default_configuration):
    from gefyra.api import list_interceptrequests

    ireqs = list_interceptrequests(config=config)
    if ireqs:
        for ireq in ireqs:
            print(ireq)
    else:
        logger.info("No active bridges found")


@stopwatch
def list_interceptrequests(config=default_configuration) -> List[str]:
    from gefyra.local.bridge import get_all_interceptrequests
    from gefyra.local.utils import (
        set_gefyra_network_from_cargo,
        set_kubeconfig_from_cargo,
    )

    # Check if kubeconfig is available through running Cargo
    config = set_kubeconfig_from_cargo(config)

    config = set_gefyra_network_from_cargo(config)
    ireqs = []
    for ireq in get_all_interceptrequests(config):
        ireqs.append(ireq["metadata"]["name"])
    return ireqs


@stopwatch
def list_containers(config=default_configuration) -> List[str]:
    from gefyra.local.bridge import get_all_containers
    from gefyra.local.utils import set_gefyra_network_from_cargo

    config = set_gefyra_network_from_cargo(config)

    return get_all_containers(config)



import io
import logging
import os
import tarfile

from docker.errors import NotFound
from docker.models.containers import Container

from gefyra.configuration import ClientConfiguration
from gefyra.local.utils import (
    build_cargo_image,
    handle_docker_create_container,
    handle_docker_remove_container,
)

logger = logging.getLogger(__name__)


def create_cargo_container(
    config: ClientConfiguration, cargo_connection_data: dict
) -> Container:
    wireguard_ip = f"{cargo_connection_data['Interface.Address']}"
    private_key = cargo_connection_data["Interface.PrivateKey"]
    dns = (
        f"{cargo_connection_data['Interface.DNS']} {config.NAMESPACE}.svc.cluster.local"
    )
    public_key = cargo_connection_data["Peer.PublicKey"]
    # docker to work with ipv4 only
    allowed_ips = cargo_connection_data["Peer.AllowedIPs"].split(",")[0]
    mtu = cargo_connection_data["MTU"]

    # build image
    image, build_logs = build_cargo_image(
        config,
        wireguard_ip=wireguard_ip,
        mtu=mtu,
        private_key=private_key,
        dns=dns,
        public_key=public_key,
        endpoint=config.CARGO_ENDPOINT,
        allowed_ips=allowed_ips,
    )
    # we only have one tag
    image_name_and_tag = image.tags[0]
    # run image
    container = handle_docker_create_container(
        config,
        image_name_and_tag,
        detach=True,
        name=config.CARGO_CONTAINER_NAME,
        auto_remove=True,
        cap_add=["NET_ADMIN"],
        privileged=True,
        volumes=["/var/run/docker.sock:/var/run/docker.sock"],
        pid_mode="host",
    )
    return container


def remove_cargo_container(config: ClientConfiguration):
    try:
        handle_docker_remove_container(config, container_id=config.CARGO_CONTAINER_NAME)
    except NotFound:
        pass


def get_cargo_ip_from_netaddress(network_address: str) -> str:
    return ".".join(network_address.split(".")[:3]) + ".149"


def get_syncdown_config(config: ClientConfiguration) -> str:
    fh = io.BytesIO()
    cargo = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
    bits, stat = cargo.get_archive("/etc/syncdown.conf")
    for chunk in bits:
        fh.write(chunk)
    fh.seek(0)
    with tarfile.open(fileobj=fh) as tf:
        fconfig = tf.extractfile(stat["name"]).read().decode()
    return fconfig


def put_syncdown_config(config: ClientConfiguration, syncdown_configuration: str):
    cargo = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
    source_f = io.BytesIO()
    source_f.write(syncdown_configuration.encode())
    fh = io.BytesIO()
    with tarfile.open(fileobj=fh, mode="w") as tf:
        info = tarfile.TarInfo("syncdown.conf")
        info.size = source_f.tell()
        source_f.seek(0)
        tf.addfile(info, fileobj=source_f)
    fh.seek(0)
    cargo.put_archive("/etc/", fh)


def delete_syncdown_job(config: ClientConfiguration, bridge_name: str):
    configfile = get_syncdown_config(config)
    old_config = configfile.split("\n")
    new_config = []
    for line in old_config:
        if line.split(";")[0] != bridge_name:
            new_config.append(line)
    configfile = "\n".join(new_config)
    put_syncdown_config(config, configfile)


def add_syncdown_job(
    config: ClientConfiguration,
    bridge_name: str,
    to_container_name: str,
    from_pod: str,
    from_container: str,
    directory: str,
):
    configfile = get_syncdown_config(config)
    relative = directory.strip("/")
    target = os.path.split(directory)[0]
    # bridge name;container name;prefix;relative directory;target directory
    configfile = (
        configfile
        + f"\n{bridge_name};{to_container_name};{from_pod}/{from_container};{relative};{target}"
    )
    put_syncdown_config(config, configfile)


def probe_wireguard_connection(config: ClientConfiguration):
    cargo = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
    for _attempt in range(0, config.CARGO_PROBE_TIMEOUT):
        _r = cargo.exec_run(f"timeout 1 ping -c 1 {config.STOWAWAY_IP}")
        if _r.exit_code != 0:
            continue
        else:
            break
    else:
        raise RuntimeError(
            f"Gefyra could not successfully confirm the Wireguard connection working. Please make sure you "
            f"are using the --endpoint argument for remote clusters and that the Kubernetes NodePort "
            f"{config.CARGO_ENDPOINT} can be reached from this machine. Please check your firewall "
            f"settings, too. If you are running a local Minikube cluster, please use the 'gefyra up --minikube' flag."
        )



import logging
from time import sleep
from typing import List, Dict

from gefyra.configuration import default_configuration

from .utils import stopwatch

logger = logging.getLogger(__name__)


def get_pods_to_intercept(
    workload_name: str, workload_type: str, namespace: str, config
) -> Dict[str, List[str]]:
    from gefyra.cluster.resources import (
        get_pods_and_containers_for_pod_name,
        get_pods_and_containers_for_workload,
    )

    pods_to_intercept = {}
    if workload_type != "pod":
        pods_to_intercept.update(
            get_pods_and_containers_for_workload(
                config, workload_name, namespace, workload_type
            )
        )
    else:
        pods_to_intercept.update(
            get_pods_and_containers_for_pod_name(config, workload_name, namespace)
        )
    return pods_to_intercept


def check_workloads(
    pods_to_intercept,
    workload_type: str,
    workload_name: str,
    container_name: str,
    namespace: str,
    config,
):
    from gefyra.cluster.resources import check_pod_valid_for_bridge

    pod_names = pods_to_intercept.keys()
    if len(pod_names) == 0:
        raise Exception("Could not find any pod to bridge.")

    cleaned_names = ["-".join(name.split("-")[:-2]) for name in pod_names]

    if workload_type != "pod" and workload_name not in cleaned_names:
        raise RuntimeError(
            f"Could not find {workload_type}/{workload_name} to bridge. Available {workload_type}:"
            f" {', '.join(cleaned_names)}"
        )
    if container_name not in [
        container for c_list in pods_to_intercept.values() for container in c_list
    ]:
        raise RuntimeError(f"Could not find container {container_name} to bridge.")

    for name in pod_names:
        check_pod_valid_for_bridge(config, name, namespace, container_name)


@stopwatch
def bridge(
    name: str,
    ports: dict,
    target: str,
    namespace: str = "default",
    sync_down_dirs: List[str] = None,
    handle_probes: bool = True,
    timeout: int = 0,
    config=default_configuration,
) -> bool:
    from docker.errors import NotFound
    from gefyra.local.utils import (
        set_gefyra_network_from_cargo,
        set_kubeconfig_from_cargo,
    )
    from gefyra.local.bridge import get_all_interceptrequests

    # Check if kubeconfig is available through running Cargo
    config = set_kubeconfig_from_cargo(config)
    config = set_gefyra_network_from_cargo(config)

    try:
        container = config.DOCKER.containers.get(name)
    except NotFound:
        logger.error(f"Could not find target container '{name}'")
        return False

    ports = [f"{key}:{value}" for key, value in ports.items()]

    try:
        local_container_ip = container.attrs["NetworkSettings"]["Networks"][
            config.NETWORK_NAME
        ]["IPAddress"]
    except KeyError:
        logger.error(
            f"The target container '{name}' is not in Gefyra's network {config.NETWORK_NAME}."
            f" Did you run 'gefyra up'?"
        )
        return False

    try:
        _bits = list(filter(None, target.split("/")))
        workload_type, workload_name = _bits[0:2]
        container_name = _bits[2] if _bits[2:] else None
    except IndexError:
        raise RuntimeError(
            "Invalid --target notation. Use <workload_type>/<workload_name>(/<container_name>)."
        )
        return False

    pods_to_intercept = get_pods_to_intercept(
        workload_name=workload_name,
        workload_type=workload_type,
        namespace=namespace,
        config=config,
    )
    if not container_name:
        container_name = pods_to_intercept[list(pods_to_intercept.keys())[0]][0]

    ireq_base_name = f"{name}-to-{namespace}.{workload_type}.{workload_name}"

    check_workloads(
        pods_to_intercept,
        workload_type=workload_type,
        workload_name=workload_name,
        container_name=container_name,
        namespace=namespace,
        config=config,
    )

    if len(pods_to_intercept.keys()) > 1:
        use_index = True
    else:
        use_index = False

    # is is required to copy at least the service account tokens from the bridged container
    if sync_down_dirs:
        sync_down_dirs = [
            "/var/run/secrets/kubernetes.io/serviceaccount"
        ] + sync_down_dirs
    else:
        sync_down_dirs = ["/var/run/secrets/kubernetes.io/serviceaccount"]

    from gefyra.local.bridge import (
        get_ireq_body,
        handle_create_interceptrequest,
    )

    from gefyra.local.cargo import add_syncdown_job

    ireqs = []
    for idx, pod in enumerate(pods_to_intercept):
        logger.info(f"Creating bridge for Pod {pod}")
        ireq_body = get_ireq_body(
            config,
            name=f"{ireq_base_name}-{idx}" if use_index else ireq_base_name,
            destination_ip=local_container_ip,
            target_pod=pod,
            target_namespace=namespace,
            target_container=container_name,
            port_mappings=ports,
            sync_down_directories=sync_down_dirs,
            handle_probes=handle_probes,
        )
        ireq = handle_create_interceptrequest(config, ireq_body, target)
        logger.debug(f"Bridge {ireq['metadata']['name']} created")
        for syncdown_dir in sync_down_dirs:
            add_syncdown_job(
                config,
                ireq["metadata"]["name"],
                name,
                pod,
                container_name,
                syncdown_dir,
            )
        ireqs.append(ireq)
    #
    # block until all bridges are in place
    #
    logger.info("Waiting for the bridge(s) to become active")

    bridges = {str(ireq["metadata"]["uid"]): False for ireq in ireqs}
    waiting_time = 0
    # timeout = 0  means no timeout
    if timeout:
        waiting_time = timeout
    while True:
        # watch whether all relevant bridges have been established
        kube_ireqs = get_all_interceptrequests(config)
        for kube_ireq in kube_ireqs:
            if kube_ireq["metadata"]["uid"] in bridges.keys() and kube_ireq.get(
                "established", False
            ):
                bridges[str(kube_ireq["metadata"]["uid"])] = True
                logger.info(
                    f"Bridge {kube_ireq['metadata']['name']} ({sum(bridges.values())}/{len(ireqs)}) established."
                )
        if all(bridges.values()):
            break
        sleep(1)
        # Raise exception in case timeout is reached
        waiting_time -= 1
        if timeout and waiting_time <= 0:
            raise RuntimeError("Timeout for bridging operation exceeded")
    logger.info("Following bridges have been established:")
    for ki in kube_ireqs:
        for port in ports:
            (
                pod_name,
                ns,
            ) = (
                ki["targetPod"],
                ki["targetNamespace"],
            )
            bridge_ports = port.split(":")
            container_port, pod_port = bridge_ports[0], bridge_ports[1]
            logger.info(
                f"Bridge for pod {pod_name} in namespace {ns} on port {pod_port} "
                f"to local container {container_name} on port {container_port}"
            )
    return True


def wait_for_deletion(ireqs: List, config=default_configuration):
    from kubernetes.watch import Watch

    w = Watch()
    deleted = []
    uids = [ireq["metadata"]["uid"] for ireq in ireqs]
    for event in w.stream(
        config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object,
        namespace=config.NAMESPACE,
        group="gefyra.dev",
        version="v1",
        plural="interceptrequests",
    ):
        if event["type"] == "DELETED":
            if event["object"]["metadata"]["uid"] in uids:
                deleted.append(event["object"]["metadata"]["uid"])
                if set(deleted) == set(uids):
                    break


@stopwatch
def unbridge(
    name: str,
    config=default_configuration,
    wait: bool = False,
) -> bool:
    from gefyra.local.bridge import handle_delete_interceptrequest
    from gefyra.local.utils import (
        set_kubeconfig_from_cargo,
    )

    config = set_kubeconfig_from_cargo(config)

    ireq = handle_delete_interceptrequest(config, name)
    if ireq:
        if wait:
            wait_for_deletion([ireq], config)
        logger.info(f"Bridge {name} removed")
    return True


@stopwatch
def unbridge_all(
    config=default_configuration,
    wait: bool = False,
) -> bool:
    from gefyra.local.bridge import (
        handle_delete_interceptrequest,
        get_all_interceptrequests,
    )
    from gefyra.local.utils import (
        set_kubeconfig_from_cargo,
    )

    config = set_kubeconfig_from_cargo(config)

    ireqs = get_all_interceptrequests(config)
    for ireq in ireqs:
        name = ireq["metadata"]["name"]
        logger.info(f"Removing Bridge {name}")
        handle_delete_interceptrequest(config, name)
    if wait:
        wait_for_deletion(config=config, ireqs=ireqs)
    return True



import logging

from gefyra.api import bridge
from gefyra.api.run import retrieve_pod_and_container, run
from gefyra.api.utils import is_port_free
from gefyra.cluster.utils import (
    get_container_command,
    get_container_image,
    get_container_ports,
    get_v1pod,
)
from gefyra.configuration import default_configuration

logger = logging.getLogger(__name__)


def _check_ports(host_ports):
    ports_not_free = []
    for port in host_ports:
        if not is_port_free(port):
            ports_not_free.append(str(port))
    if len(ports_not_free):
        raise RuntimeError(
            f"Following ports are needed for the container to run, but are occupied  \
            on your host system: {', '.join(ports_not_free)}. \
            Please provide a port mapping via --port to overwrite these ports."
        )


def reflect(
    workload: str,  # deploy/my-deployment
    namespace: str = "default",
    config=default_configuration,
    do_bridge: bool = False,
    env: list = None,
    command: str = "",
    volumes: dict = None,
    auto_remove: bool = False,
    expose_ports: bool = True,
    image: str = None,
    ports: dict = None,
):
    if expose_ports and ports:
        raise RuntimeError(
            "You cannot specify ports and expose_ports at the same time."
        )
    if not ports:
        ports = {}

    name = f"gefyra-reflect-{namespace}-{workload.replace('/', '-')}"
    pod_name, container_name = retrieve_pod_and_container(workload, namespace, config)

    pod = get_v1pod(config=config, pod_name=pod_name, namespace=namespace)
    if not image:
        image = get_container_image(pod, container_name)
    if not command:
        command = get_container_command(pod, container_name)

    if expose_ports:
        container_ports = get_container_ports(pod, container_name)
        for port in container_ports:
            host_port = ""
            if not port.host_port:
                host_port = port.container_port
            ports[port.container_port] = host_port

    host_ports = ports.values()

    _check_ports(host_ports)

    res = run(
        name=name,
        image=image,
        command=command,
        volumes=volumes,
        auto_remove=auto_remove,
        namespace=namespace,
        config=config,
        env_from=workload,
        env=env,
        detach=True,
        ports=ports,
    )

    if do_bridge:
        res = bridge(
            name=name, namespace=namespace, config=config, target=workload, ports=ports
        )
    return res



import argparse
import os
from datetime import datetime
from typing import List, Optional

from docker.models.containers import Container

from gefyra.configuration import ClientConfiguration, logger
from gefyra.local.cargoimage.cargo_dockerfile import get_dockerfile
from gefyra.local import (
    CREATED_BY_LABEL,
    ACTIVE_KUBECONFIG_LABEL,
    ACTIVE_KUBECONFIG_CONTEXT_LABEL,
    CARGO_ENDPOINT_LABEL,
    VERSION_LABEL,
)


def set_gefyra_network_from_cargo(config) -> ClientConfiguration:
    from docker.errors import NotFound

    try:
        cargo_container = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
        networks = list(cargo_container.attrs["NetworkSettings"]["Networks"].keys())
        try:
            networks.remove("bridge")
        except ValueError:
            pass
        config.NETWORK_NAME = networks[0]
    except NotFound:
        raise RuntimeError("Gefyra Cargo not running. Please run 'gefyra up' first.")
    except KeyError:
        raise RuntimeError(
            "Gefyra Cargo is not configured properly. "
            "Please set up Gefyra again with 'gefyra down' and 'gefyra up'."
        )
    return config


def get_processed_paths(base_path: str, volumes: List[str]) -> Optional[List[str]]:
    if volumes is None:
        return None
    results = []
    for volume in volumes:
        source, target = volume.rsplit(":", 1)
        if not os.path.isabs(source):
            source = os.path.realpath(os.path.join(base_path, source))
        results.append(f"{source}:{target}")
    return results


def build_cargo_image(
    config: ClientConfiguration,
    wireguard_ip: str,
    mtu: str,
    private_key: str,
    dns: str,
    public_key: str,
    endpoint: str,
    allowed_ips: str,
):
    build_args = {
        "ADDRESS": wireguard_ip,
        "MTU": str(mtu),
        "PRIVATE_KEY": private_key,
        "DNS": dns,
        "PUBLIC_KEY": public_key,
        "ENDPOINT": endpoint,
        "ALLOWED_IPS": allowed_ips,
    }
    tag = f"{config.CARGO_CONTAINER_NAME}:{datetime.now().strftime('%Y%m%d%H%M%S')}"
    # check for Cargo updates
    config.DOCKER.images.pull(config.CARGO_IMAGE)
    # build this instance
    _Dockerfile = get_dockerfile(config.CARGO_IMAGE)
    image, build_logs = config.DOCKER.images.build(
        fileobj=_Dockerfile, rm=True, forcerm=True, buildargs=build_args, tag=tag
    )
    return image, build_logs


def handle_docker_remove_container(
    config: ClientConfiguration, container: Container = None, container_id: str = None
):
    """Stop docker container, either `container` or `container_id` must be specified.

    :param config: gefyra.configuration.ClientConfiguration instance
    :param container: docker.models.containers.Container instance
    :param container_id: id or name of a docker container

    :raises AssertionError: if neither container nor container_id is specified
    :raises docker.errors.APIError: when removing of container fails
    """
    assert container or container_id, "Either container or id must be specified!"
    if not container:
        container = config.DOCKER.containers.get(container_id)

    container.remove(force=True)


def handle_docker_create_container(
    config: ClientConfiguration, image: str, **kwargs
) -> Container:
    import gefyra.configuration

    return config.DOCKER.containers.create(
        image,
        labels={
            CREATED_BY_LABEL[0]: CREATED_BY_LABEL[1],
            ACTIVE_KUBECONFIG_LABEL: config.KUBE_CONFIG_FILE,
            ACTIVE_KUBECONFIG_CONTEXT_LABEL: config.KUBE_CONTEXT,
            CARGO_ENDPOINT_LABEL: config.CARGO_ENDPOINT,
            VERSION_LABEL: gefyra.configuration.__VERSION__,
        },
        **kwargs,
    )


def handle_docker_run_container(
    config: ClientConfiguration, image: str, **kwargs
) -> Container:
    # if detach=True is in kwargs, this will return a container; otherwise the container logs (see
    # https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.ContainerCollection.run)
    # TODO: handle exception(s):
    # docker.errors.ContainerError – If the container exits with a non-zero exit code and detach is False.
    # docker.errors.ImageNotFound – If the specified image does not exist.
    # docker.errors.APIError – If the server returns an error.
    return config.DOCKER.containers.run(
        image,
        labels={
            CREATED_BY_LABEL[0]: CREATED_BY_LABEL[1],
        },
        **kwargs,
    )


def get_connection_from_kubeconfig(kubeconfig: Optional[str] = None) -> Optional[str]:
    import yaml

    if kubeconfig:
        _file = kubeconfig
    else:
        from kubernetes.config.kube_config import KUBE_CONFIG_DEFAULT_LOCATION
        from pathlib import Path

        _file = Path(KUBE_CONFIG_DEFAULT_LOCATION).expanduser()

    try:
        with open(_file, "r") as kubeconfig:
            kubecfg = yaml.safe_load(kubeconfig)
        active_ctx = next(
            filter(
                lambda x: x["name"] == kubecfg["current-context"], kubecfg["contexts"]
            )
        )
        if gefyra_connection := active_ctx.get("gefyra"):
            return gefyra_connection
        else:
            return None
    except Exception as e:  # noqa
        logger.error(
            f"Could not load Gefyra --host and --port from kubeconfig due to: {e}"
        )
        return None


def set_kubeconfig_from_cargo(config: ClientConfiguration) -> ClientConfiguration:
    from docker.errors import NotFound

    try:
        cargo_container = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
        kube_config_path = cargo_container.attrs["Config"]["Labels"][
            ACTIVE_KUBECONFIG_LABEL
        ]
        kube_context = cargo_container.attrs["Config"]["Labels"][
            ACTIVE_KUBECONFIG_CONTEXT_LABEL
        ]
        if config.KUBE_CONFIG_FILE != kube_context:
            logger.debug(
                f"Setting a different kubeconfig path from {config.KUBE_CONFIG_FILE} to {kube_config_path}"
            )
            config.KUBE_CONFIG_FILE = kube_config_path
        if config.KUBE_CONTEXT != kube_context:
            logger.debug(
                f"Setting a different kubeconfig context from {config.KUBE_CONTEXT} to {kube_context}"
            )
            config.KUBE_CONTEXT = kube_context
        return config
    except NotFound:
        logger.debug("Cargo not running, no kubeconfig set")
        return config
    except KeyError:
        logger.debug("Cargo found, but Gefyra labels missing, no kubeconfig set")
        return config


class PortMappingParser(argparse.Action):
    """Adapted from https://stackoverflow.com/questions/29986185/python-argparse-dict-arg"""

    @staticmethod
    def parse_split(split):
        # port - port
        res = {}
        if len(split) == 2:
            res[split[1]] = split[0]
            return res
        else:
            raise ValueError

    def __call__(self, parser, namespace, values, option_string=None):
        try:
            res = values.split(":")
            mapping = self.parse_split(res)
        except Exception:
            logger.error(
                "Invalid port mapping. Example valid port mapping: 8080:8081 (<ip>:host_port:container_port)."
            )
            exit(1)
        else:
            if getattr(namespace, self.dest):
                mapping.update(getattr(namespace, self.dest))
            setattr(namespace, self.dest, mapping)


class IpPortMappingParser(PortMappingParser):
    @staticmethod
    def parse_split(split):
        def v(p: str):
            if not p.isnumeric():
                raise RuntimeError(
                    f"Invalid port {p}. Please use integer numbers as port."
                )
            return p

        # port - port
        res = {}
        if len(split) == 2:
            res[v(split[1])] = v(split[0])
            return res
        elif len(split) == 3:
            res[v(split[2])] = (split[0], v(split[1]))
            return res
        else:
            raise ValueError("Invalid value for port mapping.")
